package it.unimi.di.mg4j.index;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2012 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 3 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 */


import it.unimi.dsi.fastutil.ints.IntList;
import it.unimi.dsi.fastutil.longs.LongBigList;
import it.unimi.dsi.fastutil.longs.LongList;
import it.unimi.di.mg4j.index.payload.Payload;
import it.unimi.dsi.util.ByteBufferLongBigList;
import it.unimi.dsi.util.PrefixMap;
import it.unimi.dsi.util.Properties;
import it.unimi.dsi.util.StringMap;

import java.io.IOException;

/** A {@linkplain QuasiSuccinctIndexWriter quasi-succinct} index.
 * 
 * <p>Note that the methods
 * providing {@linkplain #getPointersList() pointers}, {@linkplain #getCountsList() counts} and
 * {@linkplain #getPositionsList() positions} to index readers 
 * use reflection to detect whether the {@link LongBigList} storing
 * a component is a {@link ByteBufferLongBigList}, and 
 * in that case they return a {@linkplain ByteBufferLongBigList#copy() copy}.
 * 
 * @author Sebastiano Vigna
 */
public class QuasiSuccinctIndex extends Index {
	private static final long serialVersionUID = 1L;
	/** The default quantum. */
	public final static int DEFAULT_QUANTUM = 256;

	/** Symbolic names for additional properties of a {@link QuasiSuccinctIndex}. */
	public static enum PropertyKeys {
		/** The byte order of the index. */
		BYTEORDER,			
	}

	/** The big list of longs representing the bitstream of pointers. */
	private final LongBigList pointers;
	/** The big list of longs representing the bitstream of counts. */
	private final LongBigList counts;
	/** The big list of longs representing the bitstream of positions. */
	private final LongBigList positions;
	/** The list of offsets into {@linkplain #getPointersList() pointers}. */
	protected final LongList pointersOffsets;
	/** The list of offsets into {@linkplain #getCountsList() counts}. */
	protected final LongList countsOffsets;
	/** The list of offsets into {@linkplain #getPositionsList() positions}. */
	protected final LongList positionsOffsets;
	/** The logarithm of the skipping quantum. */
	public final int log2Quantum;

	protected QuasiSuccinctIndex( final LongBigList index, final LongBigList counts, final LongBigList positions, 
			final int numberOfDocuments, final int numberOfTerms, final long numberOfPostings, final long numberOfOccurrences, final int maxCount, final Payload payload, final int log2Quantum, final boolean hasCounts, final boolean hasPositions,
			final TermProcessor termProcessor, final String field, final Properties properties, final StringMap<? extends CharSequence> termMap, final PrefixMap<? extends CharSequence> prefixMap, 
			final IntList sizes, final LongList indexOffsets, final LongList countsOffsets, final LongList positionsOffsets ) {
		super( numberOfDocuments, numberOfTerms, numberOfPostings, numberOfOccurrences, maxCount, payload, hasCounts, hasPositions, termProcessor, field, termMap, prefixMap, sizes, properties );
		this.pointers = index;
		this.counts = counts;
		this.positions = positions;
		this.pointersOffsets = indexOffsets;
		this.countsOffsets = countsOffsets;
		this.positionsOffsets = positionsOffsets;
		this.log2Quantum = log2Quantum;
	}

	@Override
	public IndexReader getReader( int bufferSize ) throws IOException {
		return new QuasiSuccinctIndexReader( this );
	}

	public String toString() {
		return this.getClass().getSimpleName() + "[" + field + "]";
	}
	
	protected LongBigList getPointersList() {
		return pointers instanceof ByteBufferLongBigList ? ((ByteBufferLongBigList)pointers).copy() : pointers;
	}

	protected LongBigList getCountsList() {
		return counts instanceof ByteBufferLongBigList ? ((ByteBufferLongBigList)counts).copy() : counts;
	}

	protected LongBigList getPositionsList() {
		return positions instanceof ByteBufferLongBigList ? ((ByteBufferLongBigList)positions).copy() : positions;		
	}
}
