package it.unimi.di.mg4j.scratch;

import static it.unimi.di.mg4j.search.DocumentIterator.END_OF_LIST;

import it.unimi.dsi.Util;
import it.unimi.dsi.fastutil.Hash;
import it.unimi.dsi.fastutil.ints.AbstractIntComparator;
import it.unimi.dsi.fastutil.ints.IntArrays;
import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.fastutil.ints.IntOpenHashSet;
import it.unimi.dsi.fastutil.ints.IntSet;
import it.unimi.dsi.fastutil.ints.IntSets;
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap;
import it.unimi.dsi.fastutil.objects.Object2ReferenceLinkedOpenHashMap;
import it.unimi.dsi.io.FastBufferedReader;
import it.unimi.dsi.lang.MutableString;
import it.unimi.dsi.logging.ProgressLogger;
import it.unimi.di.mg4j.index.Index;
import it.unimi.di.mg4j.index.TermProcessor;
import it.unimi.di.mg4j.query.nodes.Query;
import it.unimi.di.mg4j.query.nodes.QueryBuilderVisitorException;
import it.unimi.di.mg4j.query.parser.QueryParserException;
import it.unimi.di.mg4j.query.parser.SimpleParser;
import it.unimi.di.mg4j.search.DocumentIterator;
import it.unimi.di.mg4j.search.DocumentIteratorBuilderVisitor;
import it.unimi.dsi.webgraph.BVGraph;
import it.unimi.dsi.webgraph.ImmutableGraph;
import it.unimi.dsi.webgraph.LazyIntIterator;
import it.unimi.dsi.webgraph.Transform;

import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.InvocationTargetException;
import java.net.URISyntaxException;
import java.util.Random;

import org.apache.commons.configuration.ConfigurationException;
import org.apache.log4j.Logger;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Parameter;
import com.martiansoftware.jsap.SimpleJSAP;
import com.martiansoftware.jsap.UnflaggedOption;


/** Solves a given query (in a boolean way), takes the resulting set of document pointers, and then 
 *  produces a subgraph (of a given graph whose nodes represent document pointers) made of only the 
 *  arcs stemming from a document in the result set.
 * 
 *
 */
public class GraphFromQuery {
	private static final Logger LOGGER = Util.getLogger( GraphFromQuery.class );


	/** The parser to be used. */
	private final SimpleParser parser;
	/** The graph from which subgraphs will be extracted. */
	private final ImmutableGraph graph;
	/** The transpose graph or <code>null</code>. */
	private ImmutableGraph grapht;
	/** The iterator builder visitor used to solve queries. */
	private final DocumentIteratorBuilderVisitor documentIteratorBuilderVisitor;
	/** A node permutation (meaningful only if consistent hashing is applied. */
	private final int[] nodePermutation;
	/** How many nodes should be taken (at most) from the in-neighborhood. */
	private final int maxIn;
	/** How many nodes should be taken (at most) from the out-neighborhood. */
	private final int maxOut;

	/** Creates an object ready for solving queries.
	 * 
	 * @param parser the parser to be used when solving queries.
	 * @param indexMap the map from index alias to indices.
	 * @param graph the graph whence we are going to extract subgraphs.
	 * @param grapht the transpose graph, only needed if <code>maxIn</code> or <code>maxOut</code> are positive and finite.
	 * @param maxIn how many nodes should be taken (at most) from the in-neighborhood.
	 * @param maxOut how many nodes should be taken (at most) from the out-neighborhood.
	 * @param random the random number generator to be used for selecting neighbors.
	 */
	public GraphFromQuery( final SimpleParser parser, final Object2ReferenceLinkedOpenHashMap<String,Index> indexMap, 
			final ImmutableGraph graph, final ImmutableGraph grapht, 
			final int maxIn, final int maxOut, final Random random ) {
		this.parser = parser;
		this.graph = graph;
		this.grapht = grapht;
		this.maxIn = maxIn;
		this.maxOut = maxOut;
		int n = graph.numNodes();
		
		if ( maxIn > 0 && maxIn < Integer.MAX_VALUE || maxOut > 0 && maxOut < Integer.MAX_VALUE ) {
			nodePermutation = Util.identity( n ); 
			IntArrays.shuffle( nodePermutation, random );
		} else nodePermutation = null;
		
		documentIteratorBuilderVisitor = new DocumentIteratorBuilderVisitor( indexMap, indexMap.get( parser.defaultIndex ), Integer.MAX_VALUE );
	}
	
	/** Selects a subset of size at most <code>card</code> from <code>s</code>.
	 * 
	 * @param s the set from which the subset should be extracted.
	 * @param card the maximum cardinality allowed for the extracted set.
	 * @return a subset of s selected at random.
	 */
	private IntSet select( final IntSet s, final int card ) {
		if ( card == 0 ) return IntSets.EMPTY_SET;
		else if ( card >= s.size() ) return s;
		else {
			int n = s.size();
			final int[] node = new int[ n ];
			s.toArray( node );
			IntArrays.quickSort( node, new AbstractIntComparator() {
				@Override
				public int compare( int x, int y ) {
					return nodePermutation[ x ] - nodePermutation[ y ];
				}
			});
			IntSet result = new IntOpenHashSet();
			for ( int i = 0; i < card; i++ ) result.add( node[ i ] );
			return result;
		}
	}
	
	/** Solves a query and returns the corresponding graph.
	 * 
	 * @param query the query to be resolved.
	 * @return the subgraph obtained by taking only the results of the query plus, possibly,
	 * is out/in-neighbors.
	 */
	public ImmutableGraph resolve( final Query query ) throws QueryBuilderVisitorException, IOException {
		// Solve query and prepare result set
		documentIteratorBuilderVisitor.prepare();
		DocumentIterator iterator = query.accept( documentIteratorBuilderVisitor );
		final IntSet resultSet = new IntOpenHashSet();
		for( int d; ( d = iterator.nextDocument() ) != END_OF_LIST; ) resultSet.add( d );
		iterator.dispose();
		
		// outSet
		final IntSet outSet = new IntOpenHashSet();
		final IntSet inSet = new IntOpenHashSet();
		IntIterator resultIterator = resultSet.iterator();
		while ( resultIterator.hasNext() ) {
			int node = resultIterator.nextInt(), x;
			IntSet successorSet = new IntOpenHashSet();
			if ( maxOut > 0 ) {
				LazyIntIterator successors = graph.successors( node );
				while ( ( x = successors.nextInt() ) >= 0 ) successorSet.add( x );
				outSet.addAll( select( successorSet, maxOut ) );
			}
			IntSet predecessorSet = new IntOpenHashSet();
			if ( maxIn > 0 ) {
				LazyIntIterator predecessors = grapht.successors( node );
				while ( ( x = predecessors.nextInt() ) >= 0 ) predecessorSet.add( x );
				inSet.addAll( select( predecessorSet, maxIn ) );
			}
		}
	
		// put everything together
		resultSet.addAll( outSet );
		resultSet.addAll( inSet );

		// Return the subgraph 
		return Transform.filterArcs( graph, new Transform.ArcFilter() {			
			@Override
			public boolean accept( int from, int to ) {
				return resultSet.contains( from ) && resultSet.contains( to );
			}
		} );
	}

	/** Solves a query and returns the corresponding graph.
	 * 
	 * @param queryString the query to be resolved.
	 * @return the subgraph obtained by taking only the results of the query plus, possibly,
	 * is out/in-neighbors.
	 */
	public ImmutableGraph resolve( final String queryString ) throws QueryParserException, QueryBuilderVisitorException, IOException {
		// Parse query
		return resolve( parser.parse( queryString ) );
	}

	/** Loads a set of indices of given basenames, and returns a map from index aliases to indices.
	 * 
	 * @param indexBasename the index basenames.
	 * @return the map from index alias to indices.
	 */
	public static Object2ReferenceLinkedOpenHashMap<String,Index> loadIndices( final String indexBasename[] ) throws ConfigurationException, SecurityException, IOException, URISyntaxException, ClassNotFoundException, InstantiationException, IllegalAccessException, InvocationTargetException, NoSuchMethodException {
		final Object2ReferenceLinkedOpenHashMap<String,Index> indexMap = new Object2ReferenceLinkedOpenHashMap<String,Index>( Hash.DEFAULT_INITIAL_SIZE, .5f );

		for ( String basename: indexBasename ) {
			Index index;
			
			index = Index.getInstance( new File( basename ).getPath() + "?mapped=1;sizes=1", true, true, true );
			String alias = index.field != null ? index.field : basename;
			indexMap.put( alias, index );
		}
		return indexMap;
	}

	/** Given an index map, creates and returns a parser.
	 * 
	 * @param indexMap the map from index alias to indices.
	 * @return a parser.
	 */
	public static SimpleParser getParser( final Object2ReferenceLinkedOpenHashMap<String,Index> indexMap ) {
		final Object2ObjectOpenHashMap<String,TermProcessor> termProcessors = new Object2ObjectOpenHashMap<String,TermProcessor>( indexMap.size() );
		for( String alias: indexMap.keySet() ) termProcessors.put( alias, indexMap.get( alias ).termProcessor );
		
		return new SimpleParser( indexMap.keySet(), indexMap.firstKey(), termProcessors );
	}
	
	public static void main( String[] args ) throws JSAPException, ConfigurationException, SecurityException, IOException, URISyntaxException, ClassNotFoundException, InstantiationException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, QueryBuilderVisitorException {
		SimpleJSAP jsap = new SimpleJSAP( GraphFromQuery.class.getName(), "Builds a graph based on a query. The quer(ies) are given from stdin, each preceded (tab-separated) by a query identifier.",
				new Parameter[] {
					new FlaggedOption( "out", JSAP.INTEGER_PARSER, "-1", JSAP.NOT_REQUIRED, 'o', "out", "The maximum number of outneighbors to consider (-1 for 'all')." ),
					new FlaggedOption( "in", JSAP.INTEGER_PARSER, "0", JSAP.NOT_REQUIRED, 'i', "in", "The maximum number of inneighbors to consider (-1 for 'all')." ),
					new FlaggedOption( "seed", JSAP.LONG_PARSER, "-1", JSAP.NOT_REQUIRED, 's', "seed", "The seed to be used for random number generation." ),
					new FlaggedOption( "transpose", JSAP.STRING_PARSER, null, JSAP.NOT_REQUIRED, 't', "transpose", "The basename of the transpose input graph (only needed if >=0 in-neighbors are considered)." ),
					new UnflaggedOption( "graphBasename", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.NOT_GREEDY, "The input graph basename."),
					new UnflaggedOption( "outputBasename", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.NOT_GREEDY, "The basename of the output graphs (one per query)."),
					new UnflaggedOption( "index", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.GREEDY, "The indices that we will use." )
			});

		final JSAPResult jsapResult = jsap.parse( args );
		if ( jsap.messagePrinted() ) return;

		final Object2ReferenceLinkedOpenHashMap<String,Index> indexMap;
		indexMap = loadIndices( jsapResult.getStringArray( "index" ) );
		final SimpleParser simpleParser = getParser( indexMap );
		
		int maxOut = jsapResult.getInt( "out" ); 
		if ( maxOut == -1 ) maxOut = Integer.MAX_VALUE;
		int maxIn = jsapResult.getInt( "out" );
		if ( maxIn == -1 ) maxIn = Integer.MAX_VALUE;
		
		final String basename = jsapResult.getString( "graphBasename" );
		final ImmutableGraph graph = ImmutableGraph.load( basename );
		
		final String basenamet = jsapResult.getString( "transpose" );
		if ( maxIn > 0 && basenamet == null  ) throw new IllegalArgumentException( "The transpose graph must be specified if -i is set to a value larger than 0" );
		final ImmutableGraph grapht = 
				basenamet == null ? 
						null : basenamet.equals( basename ) ? 
								graph : ImmutableGraph.load( basenamet, new ProgressLogger() ); 
		
		final long seed = jsapResult.getLong( "seed" );
		final Random random = seed == -1? new Random() : new Random( seed );
 		
		GraphFromQuery gfq = new GraphFromQuery( simpleParser, indexMap, graph, grapht, maxIn, maxOut, random );

		final String outputBasename = jsapResult.getString( "outputBasename" );

		MutableString query = new MutableString();
		FastBufferedReader fbr = new FastBufferedReader( new InputStreamReader( System.in ) );
		
		while ( fbr.readLine( query ) != null ) {
			try {
				int pos = query.indexOf( '\t' );
				LOGGER.info( query );
				String out = query.substring( 0, pos ).toString();
				ImmutableGraph result = gfq.resolve( query.substring( pos + 1 ).toString() );
				BVGraph.store( result, outputBasename + "-" + out, 
						0, // No reference
						0, // No maxrefcount
						0, // No intervals
						-1, // Default zeta-k values
						BVGraph.BLOCK_COUNT_DELTA & BVGraph.BLOCKS_DELTA & BVGraph.OFFSETS_DELTA & BVGraph.OUTDEGREES_DELTA & BVGraph.RESIDUALS_GAMMA
				);
			}
			catch ( QueryParserException e ) {
				e.printStackTrace( System.err );
				System.err.println( "Ignoring query..." );
			}
		}
	}
}
